@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2007-2008  Nokia Corporation
  *  Copyright (C) 2007-2008  Instituto Nokia de Tecnologia (INdT)
- *  Copyright (C) 2007-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2007-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -49,8 +49,8 @@
 #include "dbus.h"
 
 /* Default MTU's */
-#define RX_MTU 32767
-#define TX_MTU 32767
+#define DEFAULT_RX_MTU 32767
+#define DEFAULT_TX_MTU 32767
 
 #define TARGET_SIZE 16
 
@@ -65,6 +65,8 @@ static const guint8 PBAP_TARGET[TARGET_SIZE] = {
 /* Connection ID */
 static guint32 cid = 0x0000;
 
+static GSList *sessions = NULL;
+
 typedef struct {
 	guint8  version;
 	guint8  flags;
@@ -86,6 +88,7 @@ struct obex_commands ftp = {
 
 struct obex_commands pbap = {
 	.get		= pbap_get,
+	.setpath	= pbap_setpath,
 };
 
 static void os_reset_session(struct obex_session *os)
@@ -115,6 +118,7 @@ static void os_reset_session(struct obex_session *os)
 	os->aborted = FALSE;
 	os->offset = 0;
 	os->size = OBJECT_SIZE_DELETE;
+	os->finished = 0;
 }
 
 static void os_session_mark_aborted(struct obex_session *os)
@@ -129,11 +133,19 @@ static void os_session_mark_aborted(struct obex_session *os)
 
 static void obex_session_free(struct obex_session *os)
 {
+	sessions = g_slist_remove(sessions, os);
+
 	os_reset_session(os);
 
 	if (os->current_folder)
 		g_free(os->current_folder);
 
+	if (os->io)
+		g_io_channel_unref(os->io);
+
+	if (os->target && !memcmp(os->target, PBAP_TARGET, TARGET_SIZE))
+		pbap_phonebook_context_destroy(os);
+
 	g_free(os);
 }
 
@@ -219,7 +231,8 @@ static void cmd_connect(struct obex_session *os,
 			continue;
 
 		if (memcmp(hd.bs, FTP_TARGET, TARGET_SIZE) == 0 &&
-				os->server->services & OBEX_FTP) {
+				os->server->services &
+						(OBEX_FTP | OBEX_PCSUITE)) {
 			os->target = FTP_TARGET;
 			os->cmds = &ftp;
 			break;
@@ -229,6 +242,7 @@ static void cmd_connect(struct obex_session *os,
 				os->server->services & OBEX_PBAP) {
 			os->target = PBAP_TARGET;
 			os->cmds = &pbap;
+			pbap_phonebook_context_create(os);
 			break;
 		}
 
@@ -405,7 +419,7 @@ static void cmd_setpath(struct obex_session *os,
 		}
 
 		/* This is because OBEX_UnicodeToChar() accesses the string
-		   even if its size is zero */
+		 * even if its size is zero */
 		if (hlen == 0) {
 			os->name = g_strdup("");
 			break;
@@ -421,33 +435,39 @@ static void cmd_setpath(struct obex_session *os,
 	os->cmds->setpath(obex, obj);
 }
 
-gboolean os_prepare_get(struct obex_session *os, gchar *file, guint32 *size)
+int os_prepare_get(struct obex_session *os, gchar *file, guint32 *size)
 {
-	gint fd;
+	gint fd, err;
 	struct stat stats;
 
 	fd = open(file, O_RDONLY);
-	if (fd < 0)
+	if (fd < 0) {
+		err = -errno;
+		error("open(%s): %s (%d)", file, strerror(errno), errno);
 		goto fail;
+	}
 
-	if (fstat(fd, &stats))
+	if (fstat(fd, &stats) < 0) {
+		err = -errno;
+		error("fstat(fd=%d (%s)): %s (%d)", fd, file,
+						strerror(errno), errno);
 		goto fail;
+	}
 
 	os->fd = fd;
 	os->offset = 0;
 
 	if (stats.st_size > 0)
-		os->buf = g_new0(guint8, os->tx_mtu);
+		os->buf = g_malloc0(os->tx_mtu);
 
 	*size = stats.st_size;
 
-	return TRUE;
+	return 0;
 
 fail:
 	if (fd >= 0)
 		close(fd);
-
-	return FALSE;
+	return err;
 }
 
 static gint obex_write_stream(struct obex_session *os,
@@ -455,6 +475,7 @@ static gint obex_write_stream(struct obex_session *os,
 {
 	obex_headerdata_t hd;
 	gint32 len;
+	guint8 *ptr;
 
 	debug("obex_write_stream: name=%s type=%s tx_mtu=%d fd=%d",
 		os->name ? os->name : "", os->type ? os->type : "",
@@ -464,10 +485,11 @@ static gint obex_write_stream(struct obex_session *os,
 		return -EPERM;
 
 	if (os->fd < 0) {
-		if (os->buf == NULL)
+		if (os->buf == NULL && os->finished == FALSE)
 			return -EIO;
 
-		len = os->size - os->offset;
+		len = MIN(os->size - os->offset, os->tx_mtu);
+		ptr = os->buf + os->offset;
 		goto add_header;
 	}
 
@@ -480,7 +502,12 @@ static gint obex_write_stream(struct obex_session *os,
 		return -err;
 	}
 
+	ptr = os->buf;
+
 add_header:
+
+	hd.bs = ptr;
+
 	if (len == 0) {
 		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, 0,
 					OBEX_FL_STREAM_DATAEND);
@@ -489,12 +516,15 @@ add_header:
 		return len;
 	}
 
-	hd.bs = os->buf;
 	os->offset += len;
 
 	OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, len,
 				OBEX_FL_STREAM_DATA);
 
+	if (!memcmp(os->target, PBAP_TARGET, TARGET_SIZE))
+		if (os->offset == os->size && os->finished == FALSE)
+			OBEX_SuspendRequest(obex, obj);
+
 	return len;
 }
 
@@ -505,7 +535,7 @@ gint os_prepare_put(struct obex_session *os)
 
 	path = g_build_filename(os->current_folder, os->name, NULL);
 
-	os->fd = open(path, O_WRONLY | O_CREAT, 0600);
+	os->fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600);
 	if (os->fd < 0) {
 		error("open(%s): %s (%d)", path, strerror(errno), errno);
 		g_free(path);
@@ -529,6 +559,8 @@ gint os_prepare_put(struct obex_session *os)
 		w = write(os->fd, os->buf + len, os->offset - len);
 		if (w < 0) {
 			gint err = errno;
+			error("write(%s): %s (%d)", path, strerror(errno),
+									errno);
 			if (err == EINTR)
 				continue;
 			else
@@ -608,7 +640,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	obex_headerdata_t hd;
 	guint hlen;
 	guint8 hi;
-	guint64 free;
+	long long unsigned int free;
 	int ret;
 
 	os = OBEX_GetUserData(obex);
@@ -720,8 +752,8 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	}
 
 	free = buf.f_bsize * buf.f_bavail;
-	debug("Free space in disk: %lu", free);
-	if (os->size > free) {
+	debug("Free space in disk: %llu", free);
+	if ((guint64) os->size > free) {
 		debug("Free disk space not available");
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return FALSE;
@@ -912,7 +944,8 @@ static gboolean tty_reinit(gpointer data)
 {
 	struct server *server = data;
 
-	tty_init(server->services, server->folder, server->capability, server->devnode);
+	tty_init(server->services, server->folder, server->capability,
+					server->symlinks, server->devnode);
 
 	server_free(server);
 
@@ -923,36 +956,39 @@ static gboolean obex_handle_input(GIOChannel *io,
 				GIOCondition cond, gpointer user_data)
 {
 	obex_t *obex = user_data;
+	struct obex_session *os = OBEX_GetUserData(obex);
 
-	if (cond & G_IO_NVAL)
-		return FALSE;
-
-	if (cond & (G_IO_HUP | G_IO_ERR)) {
-		struct obex_session *os;
-
-		error("HUP");
-
-		os = OBEX_GetUserData(obex);
-		if (os->server->devnode)
-			g_idle_add(tty_reinit, os->server);
-
-		return FALSE;
+	if (cond & (G_IO_HUP | G_IO_ERR | G_IO_NVAL)) {
+		error("obex_handle_input: poll event %s%s%s",
+				(cond & G_IO_HUP) ? "HUP " : "",
+				(cond & G_IO_ERR) ? "ERR " : "",
+				(cond & G_IO_NVAL) ? "NVAL " : "");
+		goto failed;
 	}
 
 	if (OBEX_HandleInput(obex, 1) < 0) {
 		error("Handle input error");
-		return FALSE;
+		goto failed;
 	}
 
 	return TRUE;
+
+failed:
+	if (os->server->devnode) {
+		if (cond & G_IO_NVAL)
+			tty_closed();
+		else
+			g_idle_add(tty_reinit, os->server);
+	}
+
+	return FALSE;
 }
 
-gint obex_session_start(gint fd, struct server *server)
+gint obex_session_start(GIOChannel *io, struct server *server)
 {
 	struct obex_session *os;
-	GIOChannel *io;
 	obex_t *obex;
-	gint ret;
+	gint ret, fd;
 
 	os = g_new0(struct obex_session, 1);
 
@@ -963,8 +999,8 @@ gint obex_session_start(gint fd, struct server *server)
 
 	os->current_folder = g_strdup(server->folder);
 	os->server = server;
-	os->rx_mtu = RX_MTU;
-	os->tx_mtu = TX_MTU;
+	os->rx_mtu = server->rx_mtu ? server->rx_mtu : DEFAULT_RX_MTU;
+	os->tx_mtu = server->tx_mtu ? server->tx_mtu : DEFAULT_TX_MTU;
 	os->fd = -1;
 	os->size = OBJECT_SIZE_DELETE;
 
@@ -975,9 +1011,12 @@ gint obex_session_start(gint fd, struct server *server)
 	}
 
 	OBEX_SetUserData(obex, os);
+	os->obex = obex;
 
 	OBEX_SetTransportMTU(obex, os->rx_mtu, os->tx_mtu);
 
+	fd = g_io_channel_unix_get_fd(io);
+
 	ret = FdOBEX_TransportSetup(obex, fd, fd, 0);
 	if (ret < 0) {
 		obex_session_free(os);
@@ -985,17 +1024,26 @@ gint obex_session_start(gint fd, struct server *server)
 		return ret;
 	}
 
-	io = g_io_channel_unix_new(fd);
 	g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
 			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
 			obex_handle_input, obex, obex_handle_destroy);
-	g_io_channel_set_close_on_unref(io, TRUE);
-	g_io_channel_unref(io);
+	os->io = g_io_channel_ref(io);
+
+	sessions = g_slist_prepend(sessions, os);
 
 	return 0;
 }
 
-gint obex_session_stop()
+gint obex_tty_session_stop(void)
 {
+	GSList *l;
+
+	for (l = sessions; l != NULL; l = l->next) {
+		struct obex_session *os = l->data;
+
+		if (os->server->devnode && os->io)
+			g_io_channel_shutdown(os->io, TRUE, NULL);
+	}
+
 	return 0;
 }
